/*
 * Java port of Bullet (c) 2008 Martin Dvorak <jezek2@advel.cz>
 *
 * Bullet Continuous Collision Detection and Physics Library
 * Copyright (c) 2003-2008 Erwin Coumans  http://continuousphysics.com/Bullet/
 *
 * This software is provided 'as-is', without any express or implied warranty.
 * In no event will the authors be held liable for any damages arising from
 * the use of this software.
 * 
 * Permission is granted to anyone to use this software for any purpose, 
 * including commercial applications, and to alter it and redistribute it
 * freely, subject to the following restrictions:
 * 
 * 1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software. If you use this software
 *    in a product, an acknowledgment in the product documentation would be
 *    appreciated but is not required.
 * 2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 * 3. This notice may not be removed or altered from any source distribution.
 */

package com.bulletphysics.demos.concave;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.ArrayList;
import java.util.List;
import com.bulletphysics.BulletGlobals;
import com.bulletphysics.ContactAddedCallback;
import com.bulletphysics.collision.broadphase.AxisSweep3;
import com.bulletphysics.collision.broadphase.BroadphaseInterface;
import com.bulletphysics.collision.dispatch.CollisionDispatcher;
import com.bulletphysics.collision.dispatch.CollisionFlags;
import com.bulletphysics.collision.dispatch.CollisionObject;
import com.bulletphysics.collision.dispatch.DefaultCollisionConfiguration;
import com.bulletphysics.collision.narrowphase.ManifoldPoint;
import com.bulletphysics.collision.shapes.BoxShape;
import com.bulletphysics.collision.shapes.BvhTriangleMeshShape;
import com.bulletphysics.collision.shapes.CollisionShape;
import com.bulletphysics.collision.shapes.TriangleIndexVertexArray;
import com.bulletphysics.demos.opengl.DemoApplication;
import com.bulletphysics.demos.opengl.GLDebugDrawer;
import com.bulletphysics.demos.opengl.IGL;
import com.bulletphysics.demos.opengl.LWJGL;
import com.bulletphysics.dynamics.DiscreteDynamicsWorld;
import com.bulletphysics.dynamics.RigidBody;
import com.bulletphysics.dynamics.constraintsolver.ConstraintSolver;
import com.bulletphysics.dynamics.constraintsolver.SequentialImpulseConstraintSolver;
import com.bulletphysics.linearmath.Transform;
import javax.vecmath.Vector3f;
import org.lwjgl.LWJGLException;
import static com.bulletphysics.demos.opengl.IGL.*;

// JAVA TODO: update for 2.70b1

/**
 *
 * @author jezek2
 */
public class ConcaveDemo extends DemoApplication {

	// keep the collision shapes, for deletion/cleanup
	private List<CollisionShape> collisionShapes = new ArrayList<CollisionShape>();
	private TriangleIndexVertexArray indexVertexArrays;
	private BroadphaseInterface broadphase;
	private CollisionDispatcher dispatcher;
	private ConstraintSolver solver;
	private DefaultCollisionConfiguration collisionConfiguration;
	private boolean animatedMesh = false;
	
	private static ByteBuffer gVertices;
	private static ByteBuffer gIndices;
	private static BvhTriangleMeshShape trimeshShape;
	private static RigidBody staticBody;
	private static float waveheight = 5.f;

	private static final float TRIANGLE_SIZE=8.f;
	private static int NUM_VERTS_X = 30;
	private static int NUM_VERTS_Y = 30;
	private static int totalVerts = NUM_VERTS_X*NUM_VERTS_Y;

	public ConcaveDemo(IGL gl) {
		super(gl);
	}

	public void setVertexPositions(float waveheight, float offset) {
		int i;
		int j;
		Vector3f tmp = new Vector3f();

		for (i = 0; i < NUM_VERTS_X; i++) {
			for (j = 0; j < NUM_VERTS_Y; j++) {
				tmp.set(
						(i - NUM_VERTS_X * 0.5f) * TRIANGLE_SIZE,
						//0.f,
						waveheight * (float) Math.sin((float) i + offset) * (float) Math.cos((float) j + offset),
						(j - NUM_VERTS_Y * 0.5f) * TRIANGLE_SIZE);

				int index = i + j * NUM_VERTS_X;
				gVertices.putFloat((index*3 + 0) * 4, tmp.x);
				gVertices.putFloat((index*3 + 1) * 4, tmp.y);
				gVertices.putFloat((index*3 + 2) * 4, tmp.z);
			}
		}
	}

	@Override
	public void keyboardCallback(char key, int x, int y) {
		if (key == 'g') {
			animatedMesh = !animatedMesh;
			if (animatedMesh) {
				staticBody.setCollisionFlags(staticBody.getCollisionFlags() | CollisionFlags.KINEMATIC_OBJECT);
				staticBody.setActivationState(CollisionObject.DISABLE_DEACTIVATION);
			}
			else {
				staticBody.setCollisionFlags(staticBody.getCollisionFlags() & ~CollisionFlags.KINEMATIC_OBJECT);
				staticBody.forceActivationState(CollisionObject.ACTIVE_TAG);
			}
		}

		super.keyboardCallback(key, x, y);
	}

	public void initPhysics() {
		final float TRISIZE = 10f;

		BulletGlobals.gContactAddedCallback = new CustomMaterialCombinerCallback();

		//#define USE_TRIMESH_SHAPE 1
		//#ifdef USE_TRIMESH_SHAPE

		int vertStride = 3 * 4;
		int indexStride = 3 * 4;

		int totalTriangles = 2 * (NUM_VERTS_X - 1) * (NUM_VERTS_Y - 1);

		gVertices = ByteBuffer.allocateDirect(totalVerts * 3 * 4).order(ByteOrder.nativeOrder());
		gIndices = ByteBuffer.allocateDirect(totalTriangles * 3 * 4).order(ByteOrder.nativeOrder());

		int i;

		setVertexPositions(waveheight, 0.f);

		//int index=0;
		gIndices.clear();
		for (i = 0; i < NUM_VERTS_X - 1; i++) {
			for (int j = 0; j < NUM_VERTS_Y - 1; j++) {
				gIndices.putInt(j * NUM_VERTS_X + i);
				gIndices.putInt(j * NUM_VERTS_X + i + 1);
				gIndices.putInt((j + 1) * NUM_VERTS_X + i + 1);

				gIndices.putInt(j * NUM_VERTS_X + i);
				gIndices.putInt((j + 1) * NUM_VERTS_X + i + 1);
				gIndices.putInt((j + 1) * NUM_VERTS_X + i);
			}
		}
		gIndices.flip();

		indexVertexArrays = new TriangleIndexVertexArray(totalTriangles,
				gIndices,
				indexStride,
				totalVerts, gVertices, vertStride);

		boolean useQuantizedAabbCompression = true;

		//comment out the next line to read the BVH from disk (first run the demo once to create the BVH)
		//#define SERIALIZE_TO_DISK 1
		//#ifdef SERIALIZE_TO_DISK
		trimeshShape = new BvhTriangleMeshShape(indexVertexArrays, useQuantizedAabbCompression);
		collisionShapes.add(trimeshShape);

		// TODO: implement BVH serialization
//		// we can serialize the BVH data 
//		void* buffer = 0;
//		int numBytes = trimeshShape->getOptimizedBvh()->calculateSerializeBufferSize();
//		buffer = btAlignedAlloc(numBytes,16);
//		bool swapEndian = false;
//		trimeshShape->getOptimizedBvh()->serialize(buffer,numBytes,swapEndian);
//		FILE* file = fopen("bvh.bin","wb");
//		fwrite(buffer,1,numBytes,file);
//		fclose(file);
//		btAlignedFree(buffer);

//		//#else
//		trimeshShape  = new btBvhTriangleMeshShape(m_indexVertexArrays,useQuantizedAabbCompression,false);
//
//		char* fileName = "bvh.bin";
//
//		FILE* file = fopen(fileName,"rb");
//		int size=0;
//		btOptimizedBvh* bvh = 0;
//
//		if (fseek(file, 0, SEEK_END) || (size = ftell(file)) == EOF || fseek(file, 0, SEEK_SET)) {        /* File operations denied? ok, just close and return failure */
//			printf("Error: cannot get filesize from %s\n", fileName);
//			exit(0);
//		} else
//		{
//
//			fseek(file, 0, SEEK_SET);
//
//			int buffersize = size+btOptimizedBvh::getAlignmentSerializationPadding();
//
//			void* buffer = btAlignedAlloc(buffersize,16);
//			int read = fread(buffer,1,size,file);
//			fclose(file);
//			bool swapEndian = false;
//			bvh = btOptimizedBvh::deSerializeInPlace(buffer,buffersize,swapEndian);
//		}
//
//		trimeshShape->setOptimizedBvh(bvh);
//		#endif

		CollisionShape groundShape = trimeshShape;

		//#else
		//btCollisionShape* groundShape = new btBoxShape(btVector3(50,3,50));
		//m_collisionShapes.push_back(groundShape);
		//#endif //USE_TRIMESH_SHAPE

		collisionConfiguration = new DefaultCollisionConfiguration();

//		//#ifdef USE_PARALLEL_DISPATCHER
//		#ifdef USE_WIN32_THREADING
//
//		int maxNumOutstandingTasks = 4;//number of maximum outstanding tasks
//		Win32ThreadSupport* threadSupport = new Win32ThreadSupport(Win32ThreadSupport::Win32ThreadConstructionInfo(
//									"collision",
//									processCollisionTask,
//									createCollisionLocalStoreMemory,
//									maxNumOutstandingTasks));
//		#else
//		///todo other platform threading
//		///Playstation 3 SPU (SPURS)  version is available through PS3 Devnet
//		///Libspe2 SPU support will be available soon
//		///pthreads version
//		///you can hook it up to your custom task scheduler by deriving from btThreadSupportInterface
//		#endif
//
//		m_dispatcher = new	SpuGatheringCollisionDispatcher(threadSupport,maxNumOutstandingTasks,m_collisionConfiguration);
//		#else
		dispatcher = new CollisionDispatcher(collisionConfiguration);
		//#endif//USE_PARALLEL_DISPATCHER

		Vector3f worldMin = new Vector3f(-1000f, -1000f, -1000f);
		Vector3f worldMax = new Vector3f(1000f, 1000f, 1000f);
		broadphase = new AxisSweep3(worldMin, worldMax);
		//broadphase = new SimpleBroadphase();
		solver = new SequentialImpulseConstraintSolver();
		dynamicsWorld = new DiscreteDynamicsWorld(dispatcher, broadphase, solver, collisionConfiguration);
		//#ifdef USE_PARALLEL_DISPATCHER
		//m_dynamicsWorld->getDispatchInfo().m_enableSPU=true;
		//#endif //USE_PARALLEL_DISPATCHER

		// JAVA NOTE: added
		dynamicsWorld.setDebugDrawer(new GLDebugDrawer(gl));
		
		float mass = 0f;
		Transform startTransform = new Transform();
		startTransform.setIdentity();
		startTransform.origin.set(0f, -2f, 0f);

		CollisionShape colShape = new BoxShape(new Vector3f(1f, 1f, 1f));
		collisionShapes.add(colShape);

		{
			for (i = 0; i < 10; i++) {
				//btCollisionShape* colShape = new btCapsuleShape(0.5,2.0);//boxShape = new btSphereShape(1.f);
				startTransform.origin.set(2f * i, 10f, 1f);
				localCreateRigidBody(1f, startTransform, colShape);
			}
		}

		startTransform.setIdentity();
		staticBody = localCreateRigidBody(mass, startTransform, groundShape);

		staticBody.setCollisionFlags(staticBody.getCollisionFlags() | CollisionFlags.STATIC_OBJECT);

		// enable custom material callback
		staticBody.setCollisionFlags(staticBody.getCollisionFlags() | CollisionFlags.CUSTOM_MATERIAL_CALLBACK);
	}

	private static float offset = 0f;
	
	@Override
	public void clientMoveAndDisplay() {
		gl.glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

		float dt = getDeltaTimeMicroseconds() * 0.000001f;

		if (animatedMesh) {
			long t0 = System.nanoTime();
			
			offset += 0.01f;

			setVertexPositions(waveheight, offset);

			// JAVA NOTE: 2.70b1: replace with proper code
			trimeshShape.refitTree(null, null);

			// clear all contact points involving mesh proxy. Note: this is a slow/unoptimized operation.
			dynamicsWorld.getBroadphase().getOverlappingPairCache().cleanProxyFromPairs(staticBody.getBroadphaseHandle(), getDynamicsWorld().getDispatcher());
			
			BulletGlobals.updateTime = (System.nanoTime() - t0) / 1000000;
		}

		dynamicsWorld.stepSimulation(dt);

		// optional but useful: debug drawing
		dynamicsWorld.debugDrawWorld();

		renderme();

		//glFlush();
		//glutSwapBuffers();
	}

	@Override
	public void displayCallback() {
		gl.glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

		renderme();

		// optional but useful: debug drawing
		if (dynamicsWorld != null) {
			dynamicsWorld.debugDrawWorld();
		}
		
		//glFlush();
		//glutSwapBuffers();
	}

	/**
	 * User can override this material combiner by implementing gContactAddedCallback
	 * and setting body0->m_collisionFlags |= btCollisionObject::customMaterialCallback
	 */
	private static float calculateCombinedFriction(float friction0, float friction1) {
		float friction = friction0 * friction1;

		float MAX_FRICTION = 10f;
		if (friction < -MAX_FRICTION) {
			friction = -MAX_FRICTION;
		}
		if (friction > MAX_FRICTION) {
			friction = MAX_FRICTION;
		}
		return friction;
	}

	private static float calculateCombinedRestitution(float restitution0, float restitution1) {
		return restitution0 * restitution1;
	}
	
	private static class CustomMaterialCombinerCallback implements ContactAddedCallback {
		public boolean invoke(ManifoldPoint cp, CollisionObject colObj0, int partId0, int index0, CollisionObject colObj1, int partId1, int index1) {
			float friction0 = colObj0.getFriction();
			float friction1 = colObj1.getFriction();
			float restitution0 = colObj0.getRestitution();
			float restitution1 = colObj1.getRestitution();

			if ((colObj0.getCollisionFlags() & CollisionFlags.CUSTOM_MATERIAL_CALLBACK) != 0) {
				friction0 = 1f; //partId0,index0
				restitution0 = 0f;
			}
			if ((colObj1.getCollisionFlags() & CollisionFlags.CUSTOM_MATERIAL_CALLBACK) != 0) {
				if ((index1 & 1) != 0) {
					friction1 = 1f; //partId1,index1
				}
				else {
					friction1 = 0f;
				}
				restitution1 = 0f;
			}

			cp.combinedFriction = calculateCombinedFriction(friction0, friction1);
			cp.combinedRestitution = calculateCombinedRestitution(restitution0, restitution1);

			// this return value is currently ignored, but to be on the safe side: return false if you don't calculate friction
			return true;
		}
	}
	
	public static void main(String[] args) throws LWJGLException {
		ConcaveDemo concaveDemo = new ConcaveDemo(LWJGL.getGL());
		concaveDemo.initPhysics();
		concaveDemo.setCameraDistance(30f);

		LWJGL.main(args, 800, 600, "Static Concave Mesh Demo", concaveDemo);
	}
	
}
